"""Generate the test code for the ``Descend`` methods and ``VisitorThrough``."""

import io
import os
import pathlib
import sys
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.python.common
import aas_core_codegen.python.naming
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
)

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    this_path = pathlib.Path(os.path.realpath(__file__))

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    # noinspection PyListCreation
    blocks = [
        warning,
        Stripped(
            f"""\
\"\"\"
Test jointly :py:method:`{aas_module}.types.Class.descend` and
:py:method:`{aas_module}.types.PassThroughVisitor`.
\"\"\""""
        ),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
from typing import (
{I}List,
{I}Sequence,
)
import unittest

import {aas_module}.types as aas_types

import tests.common
import tests.common_jsonization"""
        ),
        Stripped(
            f"""\
class _TracingVisitor(aas_types.PassThroughVisitor):
{I}\"\"\"Visit the instances and trace them.\"\"\"

{I}def __init__(self) -> None:
{II}\"\"\"Initialize with an empty log.\"\"\"
{II}self._log = []  # type: List[str]

{I}@property
{I}def log(self) -> Sequence[str]:
{II}\"\"\"Get the tracing log.\"\"\"
{II}return self._log

{I}def visit(self, that: aas_types.Class) -> None:
{II}self._log.append(
{III}tests.common.trace(that)
{II})
{II}super().visit(that)"""
        ),
        Stripped(
            f"""\
def assert_tracing_logs_from_descend_and_visitor_are_the_same(
{II}that: aas_types.Class,
{II}test_case: unittest.TestCase
) -> None:
{I}\"\"\"
{I}Check that the tracing logs are the same when :paramref:`that` instance
{I}is visited and when :paramref:`that` is ran through
{I}:py:method:`aas_types.Class.descend`.

{I}:param that: instance to be iterated over
{I}:param test_case: in which this assertion runs
{I}:raise: :py:class:`AssertionError` if the logs differ
{I}\"\"\"
{I}log_from_descend = [
{II}tests.common.trace(something)
{II}for something in that.descend()
{I}]

{I}visitor = _TracingVisitor()
{I}visitor.visit(that)
{I}log_from_visitor = visitor.log

{I}test_case.assertGreater(len(log_from_visitor), 0)
{I}test_case.assertEqual(
{II}tests.common.trace(that),
{II}log_from_visitor[0]
{I})

{I}# noinspection PyTypeChecker
{I}test_case.assertListEqual(log_from_descend, log_from_visitor[1:])  # type: ignore"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        cls_name = aas_core_codegen.python.naming.class_name(our_type.name)

        model_type_literal = aas_core_codegen.python.common.string_literal(
            aas_core_codegen.naming.json_model_type(our_type.name)
        )

        load_maximal_name = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"load_maximal_{our_type.name}")
        )

        blocks.append(
            Stripped(
                f"""\
class Test_{cls_name}(unittest.TestCase):
{I}def test_descend_against_recorded_trace_log(self) -> None:
{II}instance = tests.common_jsonization.{load_maximal_name}()
{II}expected_path = (
{IIII}tests.common.TEST_DATA_DIR
{IIII}/ "descend_and_pass_through_visitor"
{IIII}/ {model_type_literal}
{IIII}/ "maximal.json.trace"
{II})

{II}log = [
{III}tests.common.trace(something)
{III}for something in instance.descend()
{II}]
{II}got_text = tests.common.trace_log_as_text_file_content(log)
{II}tests.common.record_or_check(expected_path, got_text)

{I}def test_descend_against_visitor(self) -> None:
{II}instance = tests.common_jsonization.{load_maximal_name}()
{II}assert_tracing_logs_from_descend_and_visitor_are_the_same(
{III}instance,
{III}self
{II})"""
            )
        )

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT
        / "tests/test_descend_and_pass_through_visitor.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Provide functionality used across development scripts."""

import io
import os
import pathlib

from aas_core_codegen.common import Stripped
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.run
import aas_core_codegen.python.common
from aas_core_codegen import intermediate
import aas_core_meta.v3

from icontract import require

import aas_core3

AAS_MODULE = aas_core_codegen.python.common.QualifiedModuleName(aas_core3.__name__)


REPO_ROOT = pathlib.Path(os.path.realpath(__file__)).parent.parent.parent


def load_symbol_table() -> intermediate.SymbolTable:
    """Load the symbol table from the meta-model."""
    model_path = pathlib.Path(aas_core_meta.v3.__file__)
    assert model_path.exists() and model_path.is_file(), model_path

    text = model_path.read_text(encoding="utf-8")

    atok, parse_exception = aas_core_codegen.parse.source_to_atok(source=text)
    if parse_exception:
        if isinstance(parse_exception, SyntaxError):
            raise RuntimeError(
                f"Failed to parse the meta-model {model_path}: "
                f"invalid syntax at line {parse_exception.lineno}\n"
            )
        else:
            raise RuntimeError(
                f"Failed to parse the meta-model {model_path}: " f"{parse_exception}\n"
            )

    assert atok is not None

    import_errors = aas_core_codegen.parse.check_expected_imports(atok=atok)
    if import_errors:
        writer = io.StringIO()
        aas_core_codegen.run.write_error_report(
            message="One or more unexpected imports in the meta-model",
            errors=import_errors,
            stderr=writer,
        )

        raise RuntimeError(writer.getvalue())

    lineno_columner = aas_core_codegen.common.LinenoColumner(atok=atok)

    parsed_symbol_table, error = aas_core_codegen.parse.atok_to_symbol_table(atok=atok)
    if error is not None:
        writer = io.StringIO()
        aas_core_codegen.run.write_error_report(
            message=f"Failed to construct the symbol table from {model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )

        raise RuntimeError(writer.getvalue())

    assert parsed_symbol_table is not None

    ir_symbol_table, error = intermediate.translate(
        parsed_symbol_table=parsed_symbol_table,
        atok=atok,
    )
    if error is not None:
        writer = io.StringIO()
        aas_core_codegen.run.write_error_report(
            message=f"Failed to translate the parsed symbol table "
            f"to intermediate symbol table "
            f"based on {model_path}",
            errors=[lineno_columner.error_message(error)],
            stderr=writer,
        )

        raise RuntimeError(writer.getvalue())

    assert ir_symbol_table is not None

    return ir_symbol_table


@require(lambda relative_path: not relative_path.is_absolute())
def generate_warning_comment(relative_path: pathlib.Path) -> Stripped:
    """Generate the warning that the file has been automatically generated."""
    # NOTE (mristin, 2022-10-20):
    # We use POSIX path here to be consistent regardless of the execution environment.
    posix_relative_path = pathlib.PurePosixPath(relative_path)
    return Stripped(
        f"""\
# This code has been automatically generated by:
# {posix_relative_path}
# Do NOT edit or append."""
    )



"""Generate the test code for the jsonization of classes outside a container."""

import io
import os
import pathlib
import sys
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.python.common
import aas_core_codegen.python.naming
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
    INDENT4 as IIII,
    INDENT5 as IIIII,
)

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    this_path = pathlib.Path(os.path.realpath(__file__))

    test_data_dir = dev_scripts.test_codegen.common.REPO_ROOT / "test_data"

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    blocks = [
        warning,
        Stripped(
            '"""Test de/serialization from JSON of concrete classes '
            'outside a container."""'
        ),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
import unittest

import {aas_module}.jsonization as aas_jsonization

import tests.common
import tests.common_jsonization"""
        ),
    ]  # type: List[Stripped]

    environment_cls = symbol_table.must_find_concrete_class(
        aas_core_codegen.common.Identifier("Environment")
    )

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        container_cls = dev_scripts.test_codegen.test_data_io.determine_container_class(
            cls=our_type, test_data_dir=test_data_dir, environment_cls=environment_cls
        )

        if container_cls is our_type:
            # NOTE (mristin, 2022-10-23):
            # These classes are tested already in
            # :py:module:`tests.test_jsonization_of_concrete_classes`. We only need to
            # test here for class instances contained in a container.
            continue

        cls_name = aas_core_codegen.python.naming.class_name(our_type.name)

        load_maximal_name = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"load_maximal_{our_type.name}")
        )

        from_jsonable_name = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"{our_type.name}_from_jsonable")
        )

        blocks.append(
            Stripped(
                f"""\
class Test_{cls_name}(unittest.TestCase):
{I}def test_round_trip(self) -> None:
{II}instance = tests.common_jsonization.{load_maximal_name}()

{II}jsonable = aas_jsonization.to_jsonable(instance)

{II}another_instance = aas_jsonization.{from_jsonable_name}(
{III}jsonable
{II})

{II}another_jsonable = aas_jsonization.to_jsonable(another_instance)

{II}# Check the round-trip
{II}self.assertListEqual(
{III}[],
{III}list(
{IIII}map(
{IIIII}str,
{IIIII}tests.common_jsonization.check_equal(jsonable, another_jsonable)
{IIII})
{III})
{II})"""
            )
        )

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT
        / "tests/test_jsonization_of_concrete_classes_outside_container.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Generate the test code for the ``OverXOrEmpty`` methods."""

import io
import os
import pathlib
import sys
import textwrap
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python import naming as python_naming
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
    INDENT3 as III,
)

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    this_path = pathlib.Path(os.path.realpath(__file__))

    test_data_dir = dev_scripts.test_codegen.common.REPO_ROOT / "test_data"

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    blocks = [
        warning,
        Stripped('"""Test ``over_X_or_empty`` methods."""'),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            """\
import unittest

import tests.common
import tests.common_jsonization"""
        ),
    ]  # type: List[Stripped]

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        test_case_methods = []  # type: List[Stripped]

        cls_name = python_naming.class_name(our_type.name)

        model_type_literal = aas_core_codegen.python.common.string_literal(
            aas_core_codegen.naming.json_model_type(our_type.name)
        )

        load_maximal_name = python_naming.function_name(
            aas_core_codegen.common.Identifier(f"load_maximal_{our_type.name}")
        )

        load_minimal_name = python_naming.function_name(
            aas_core_codegen.common.Identifier(f"load_minimal_{our_type.name}")
        )

        maximal_instance = dev_scripts.test_codegen.test_data_io.load_maximal(
            test_data_dir=test_data_dir, cls=our_type
        )

        minimal_instance = dev_scripts.test_codegen.test_data_io.load_minimal(
            test_data_dir=test_data_dir, cls=our_type
        )

        for prop in our_type.properties:
            if isinstance(
                prop.type_annotation, intermediate.OptionalTypeAnnotation
            ) and isinstance(
                prop.type_annotation.value, intermediate.ListTypeAnnotation
            ):
                over_x_or_empty_name = python_naming.method_name(
                    aas_core_codegen.common.Identifier(f"over_{prop.name}_or_empty")
                )

                test_case_method_name = python_naming.method_name(
                    aas_core_codegen.common.Identifier(
                        f"test_{over_x_or_empty_name}_"
                        f"on_maximal_instance_against_recorded"
                    )
                )

                prop_name = python_naming.property_name(prop.name)

                file_name = f"{over_x_or_empty_name}.trace"

                if getattr(maximal_instance, prop_name) is not None:
                    test_case_methods.append(
                        Stripped(
                            f"""\
def {test_case_method_name}(self) -> None:
{I}instance = tests.common_jsonization.{load_maximal_name}()
{I}expected_path = (
{III}tests.common.TEST_DATA_DIR
{III}/ "test_over_X_or_empty"
{III}/ {model_type_literal}
{III}/ "on_maximal.json"
{III}/ {aas_core_codegen.python.common.string_literal(file_name)}
{I})

{I}assert instance.{prop_name} is not None, (
{II}"Unexpected {prop_name} is None "
{II}"in the maximal example of "
{II}"{cls_name}"
{I})

{I}log = [
{II}tests.common.trace(
{III}list(instance.{over_x_or_empty_name}())
{II})
{I}]
{I}got_text = tests.common.trace_log_as_text_file_content(log)
{I}tests.common.record_or_check(expected_path, got_text)"""
                        )
                    )
                else:
                    test_case_methods.append(
                        Stripped(
                            f"""\
# The maximal example of {cls_name} contains no {prop_name},
# so we can not generate the corresponding test case
# {test_case_method_name}."""
                        )
                    )

                test_case_method_name = python_naming.method_name(
                    aas_core_codegen.common.Identifier(
                        f"test_{over_x_or_empty_name}_"
                        f"on_minimal_instance_against_recorded"
                    )
                )

                if not getattr(minimal_instance, prop_name) is not None:
                    test_case_methods.append(
                        Stripped(
                            f"""\
def {test_case_method_name}(self) -> None:
{I}instance = tests.common_jsonization.{load_minimal_name}()
{I}expected_path = (
{III}tests.common.TEST_DATA_DIR
{III}/ "test_over_X_or_empty"
{III}/ {model_type_literal}
{III}/ "on_minimal.json"
{III}/ {aas_core_codegen.python.common.string_literal(file_name)}
{I})

{I}assert instance.{prop_name} is None, (
{II}"Unexpected {prop_name} not None "
{II}"in the minimal example of "
{II}"{cls_name}"
{I})

{I}log = [
{II}tests.common.trace(
{III}list(instance.{over_x_or_empty_name}())
{II})
{I}]
{I}got_text = tests.common.trace_log_as_text_file_content(log)
{I}tests.common.record_or_check(expected_path, got_text)"""
                        )
                    )
                else:
                    test_case_methods.append(
                        Stripped(
                            f"""\
# The minimal example of {cls_name} contains no {prop_name},
# so we can not generate the corresponding test case
# {test_case_method_name}."""
                        )
                    )

        if len(test_case_methods) > 0:
            class_writer = io.StringIO()
            class_writer.write(
                f"""\
class Test_{cls_name}(unittest.TestCase):
{I}\"\"\"
{I}Test ``over_X_or_default`` on instances of
{I}:py:class:`{aas_module}.types.{cls_name}`.
{I}\"\"\""""
            )

            for test_case_method in test_case_methods:
                class_writer.write("\n\n")
                class_writer.write(textwrap.indent(test_case_method, I))

            blocks.append(Stripped(class_writer.getvalue()))

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()

    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT / "tests/test_over_x_or_empty.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Generate the test code for the xmlization of classes outside a container."""

import io
import os
import pathlib
import sys
import textwrap
from typing import List

import aas_core_codegen
import aas_core_codegen.common
import aas_core_codegen.naming
import aas_core_codegen.parse
import aas_core_codegen.python.common
import aas_core_codegen.python.naming
import aas_core_codegen.run
from aas_core_codegen import intermediate
from aas_core_codegen.common import Stripped
from aas_core_codegen.python.common import (
    INDENT as I,
    INDENT2 as II,
)

import dev_scripts.test_codegen.common
import dev_scripts.test_codegen.test_data_io


def main() -> int:
    """Execute the main routine."""
    symbol_table = dev_scripts.test_codegen.common.load_symbol_table()

    this_path = pathlib.Path(os.path.realpath(__file__))

    test_data_dir = dev_scripts.test_codegen.common.REPO_ROOT / "test_data"

    warning = dev_scripts.test_codegen.common.generate_warning_comment(
        this_path.relative_to(dev_scripts.test_codegen.common.REPO_ROOT)
    )

    aas_module = dev_scripts.test_codegen.common.AAS_MODULE

    blocks = [
        warning,
        Stripped(
            '"""Test de/serialization from XML of concrete classes '
            'outside a container."""'
        ),
        Stripped("# pylint: disable=missing-docstring"),
        Stripped(
            f"""\
import unittest
import xml.etree.ElementTree as ET

import {aas_module}.xmlization as aas_xmlization
import {aas_module}.verification as aas_verification

import tests.common
import tests.common_jsonization
import tests.common_xmlization"""
        ),
    ]  # type: List[Stripped]

    environment_cls = symbol_table.must_find_concrete_class(
        aas_core_codegen.common.Identifier("Environment")
    )

    for our_type in symbol_table.our_types:
        if not isinstance(our_type, intermediate.ConcreteClass):
            continue

        container_cls = dev_scripts.test_codegen.test_data_io.determine_container_class(
            cls=our_type, test_data_dir=test_data_dir, environment_cls=environment_cls
        )

        if container_cls is our_type:
            # NOTE (mristin, 2022-06-27):
            # These classes are tested already in
            # ``test_xmlization_of_concrete_classes``. We only need to test for class
            # instances contained in a container.
            continue

        assert container_cls is environment_cls

        load_maximal_name = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"load_maximal_{our_type.name}")
        )

        from_str = aas_core_codegen.python.naming.function_name(
            aas_core_codegen.common.Identifier(f"{our_type.name}_from_str")
        )

        methods = [
            Stripped(
                f"""\
def test_ok(self) -> None:
{I}instance = tests.common_jsonization.{load_maximal_name}()
{I}text = aas_xmlization.to_str(instance)

{I}another_instance = aas_xmlization.{from_str}(
{II}text
{I})

{I}errors = list(aas_verification.verify(another_instance))
{I}self.assertListEqual([], errors)

{I}et_instance = ET.fromstring(text)
{I}tests.common_xmlization.remove_redundant_whitespace(et_instance)

{I}et_another_instance = ET.fromstring(
{II}aas_xmlization.to_str(another_instance)
{I})
{I}tests.common_xmlization.remove_redundant_whitespace(et_another_instance)

{I}tests.common_xmlization.assert_elements_equal(
{II}et_instance, et_another_instance
{I})"""
            )
        ]  # type: List[Stripped]

        cls_name = aas_core_codegen.python.naming.class_name(our_type.name)

        class_writer = io.StringIO()
        class_writer.write(
            f"""\
class Test_{cls_name}(unittest.TestCase):
{I}\"\"\"
{I}Test XML de/serialization of the concrete class
{I}:py:class:`{aas_module}.types.{cls_name}` outside a container.
{I}\"\"\""""
        )

        for method in methods:
            class_writer.write("\n\n")
            class_writer.write(textwrap.indent(method, I))

        blocks.append(Stripped(class_writer.getvalue()))

    blocks.append(
        Stripped(
            """\
if __name__ == "__main__":
    unittest.main()"""
        )
    )

    blocks.append(warning)

    writer = io.StringIO()
    for i, block in enumerate(blocks):
        if i > 0:
            writer.write("\n\n\n")

        writer.write(block)

    writer.write("\n")

    target_pth = (
        dev_scripts.test_codegen.common.REPO_ROOT
        / "tests/test_xmlization_of_concrete_classes_outside_container.py"
    )
    target_pth.write_text(writer.getvalue(), encoding="utf-8")

    return 0


if __name__ == "__main__":
    sys.exit(main())



"""Read test data."""
import inspect
import itertools
import json
import pathlib

import aas_core_codegen.naming
from aas_core_codegen import intermediate
from aas_core_codegen.common import Identifier
from aas_core_codegen.python import naming as python_naming
from icontract import require

from aas_core3 import types as aas_types, jsonization as aas_jsonization


def _load_instance_from_environment(
    path: pathlib.Path, cls: intermediate.ConcreteClass
) -> aas_types.Class:
    """Load an instance from the environment at ``path``."""
    with path.open("rt", encoding="utf-8") as fid:
        jsonable = json.load(fid)

    container = aas_jsonization.environment_from_jsonable(jsonable)

    cls_name_python = python_naming.class_name(cls.name)
    cls_python = getattr(aas_types, cls_name_python, None)
    assert cls_python is not None, (
        f"Could not find the class with name {cls_name_python!r} "
        f"in {aas_types.__name__}"
    )

    assert isinstance(cls_python, type)
    assert inspect.isclass(cls_python)

    for something in itertools.chain([container], container.descend()):
        if isinstance(something, cls_python):
            assert isinstance(something, aas_types.Class)
            return something

    raise AssertionError(
        f"Failed to find an instance of {cls_name_python!r} "
        f"beneath and including {container}"
    )


def _load_self_contained_instance(
    path: pathlib.Path, cls: intermediate.ConcreteClass
) -> aas_types.Class:
    """Load an instance self-contained at ``path``."""
    with path.open("rt", encoding="utf-8") as fid:
        jsonable = json.load(fid)

    from_function_name = python_naming.function_name(
        Identifier(f"{cls.name}_from_jsonable")
    )
    from_function = getattr(aas_jsonization, from_function_name, None)
    if from_function is None:
        raise AssertionError(
            f"Function {from_function_name!r} not found "
            f"in {aas_jsonization.__name__}"
        )

    assert inspect.isfunction(from_function)
    instance = from_function(jsonable)

    cls_name_python = python_naming.class_name(cls.name)
    cls_python = getattr(aas_types, cls_name_python, None)
    assert cls_python is not None, (
        f"Could not find the class with name {cls_name_python!r} "
        f"in {aas_types.__name__}"
    )

    assert isinstance(cls_python, type)
    assert inspect.isclass(cls_python)

    assert isinstance(
        instance, cls_python
    ), f"Expected an instance of {cls_python} at {path}, but got: {instance}"

    assert isinstance(instance, aas_types.Class)

    return instance


def load_maximal(
    test_data_dir: pathlib.Path, cls: intermediate.ConcreteClass
) -> aas_types.Class:
    """Load the maximal example for the given class."""
    cls_name_json = aas_core_codegen.naming.json_model_type(cls.name)

    base_path = (
        test_data_dir / "Json" / "ContainedInEnvironment" / "Expected" / cls_name_json
    )
    if base_path.exists():
        maximal_path = base_path / "maximal.json"

        return _load_instance_from_environment(path=maximal_path, cls=cls)

    base_path = test_data_dir / "Json" / "SelfContained" / "Expected" / cls_name_json
    if base_path.exists():
        maximal_path = base_path / "maximal.json"
        return _load_self_contained_instance(path=maximal_path, cls=cls)

    raise FileNotFoundError(
        f"The maximal JSON file for {cls.name!r} could not be found"
    )


def load_minimal(
    test_data_dir: pathlib.Path, cls: intermediate.ConcreteClass
) -> aas_types.Class:
    """Load the minimal example for the given class."""
    cls_name_json = aas_core_codegen.naming.json_model_type(cls.name)

    base_path = (
        test_data_dir / "Json" / "ContainedInEnvironment" / "Expected" / cls_name_json
    )
    if base_path.exists():
        minimal_path = base_path / "minimal.json"

        return _load_instance_from_environment(path=minimal_path, cls=cls)

    base_path = test_data_dir / "Json" / "SelfContained" / "Expected" / cls_name_json
    if base_path.exists():
        minimal_path = base_path / "minimal.json"
        return _load_self_contained_instance(path=minimal_path, cls=cls)

    raise FileNotFoundError(
        f"The minimal JSON file for {cls.name!r} could not be found"
    )


@require(lambda test_data_dir: test_data_dir.is_dir())
@require(lambda environment_cls: environment_cls.name == "Environment")
def determine_container_class(
    cls: intermediate.ConcreteClass,
    test_data_dir: pathlib.Path,
    environment_cls: intermediate.ConcreteClass,
) -> intermediate.ConcreteClass:
    """Determine the container class based on the test data directory."""
    cls_name_json = aas_core_codegen.naming.json_model_type(cls.name)

    contained_in_environment_path = (
        test_data_dir / "Json" / "ContainedInEnvironment" / "Expected" / cls_name_json
    )

    self_contained_path = (
        test_data_dir / "Json" / "SelfContained" / "Expected" / cls_name_json
    )

    if contained_in_environment_path.exists():
        return environment_cls
    elif self_contained_path.exists():
        return cls
    else:
        raise RuntimeError(
            f"Neither {contained_in_environment_path} nor {self_contained_path} "
            f"exist beneath {test_data_dir}. We do not know how to infer "
            f"the kind of how the instance of {cls_name_json} is contained."
        )
